/*GPL*START*
 * 
 * simpleppp - a simple ppp dialer and monitor
 * 
 * Copyright (C) 1999 by Johannes Overmann <overmann@iname.com>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * *GPL*END*/  

#include <sys/types.h>
#include <sys/time.h>
#include <gtk/gtk.h>
#include <stdlib.h>
#include "tappconfig.h"

// history:
// 1999:
// 23 Aug 21:12 started (first gtk app)
// 24 Aug 00:04 gui looks good
// 20 Sep 21:12 v0.0.2 80% done, usable since 2 weeks
// 22 Sep 23:33 v0.0.3 y grid fixed, looks good now, 5 grid layers (strong,weak,w2,w3,w4)
// 23 Sep 19:59 v0.0.4 x grid started, 
// 23 Sep 22:37 v0.0.4 x grid done, looks good scaling values, 5 levels
// 27 Sep 19:13 v0.1.0 int overflow for x grid fixed (timescale limited to 2^31 values)
// 28 Sep 00:50 v0.1.1 total time/costs implemented
// 29 Sep 19:10 v0.1.2 total time/costs fixed
// 29 Sep 21:43 v0.1.3 some timescale bugs fixed, release to dist
// 05 Feb 01:55 v0.1.4 more timeticks




#define Tsec_min   (60.0)
#define Tsec_hour  (60.0*         Tsec_min)
#define Tsec_day   (24.0*         Tsec_hour)
#define Tsec_week  (7.0*          Tsec_day)
#define Tsec_month (365.2425/12.0*Tsec_day)
#define Tsec_year  (365.2425*     Tsec_day)

double time_scale_tab[] = {
   0.0001, 0.0002, 0.0005, 0.001, 0.002, 0.005, 0.01, 0.02, 0.05, 0.1, 0.2, 0.5,
   1, 2, 5, 10, 15, 20, 30, 45, // sec
   1*Tsec_min, 1.5*Tsec_min, 2*Tsec_min, 3*Tsec_min, 4*Tsec_min, 5*Tsec_min, 7*Tsec_min, 10*Tsec_min, 15*Tsec_min, 20*Tsec_min, 30*Tsec_min, 45*Tsec_min,
   1*Tsec_hour, 1.5*Tsec_hour, 2*Tsec_hour, 3*Tsec_hour, 4*Tsec_hour, 5*Tsec_hour, 6*Tsec_hour, 8*Tsec_hour, 12*Tsec_hour,
   1*Tsec_day, 2*Tsec_day, 3*Tsec_day, 4*Tsec_day, 5*Tsec_day, 6*Tsec_day, 
   1*Tsec_week, 2*Tsec_week, 3*Tsec_week, 
   1*Tsec_month, 2*Tsec_month, 3*Tsec_month, 4*Tsec_month, 6*Tsec_month,
   1*Tsec_year, 2*Tsec_year, 5*Tsec_year, 10*Tsec_year, 20*Tsec_year, 50.0*Tsec_year,
   100.0*Tsec_year, 200.0*Tsec_year, 500.0*Tsec_year, 
   1000.0*Tsec_year, 2000.0*Tsec_year, 5000.0*Tsec_year, 
   10000.0*Tsec_year, 20000.0*Tsec_year, 50000.0*Tsec_year, 
   100000.0*Tsec_year, 200000.0*Tsec_year, 500000.0*Tsec_year, 1000000.0*Tsec_year, 2000000.0*Tsec_year,
   0
};
int time_grid_skip[] = {
   -0, -1, -2, -1, -1, -2, -1, -1, -2, -1, -1, -2, // msec
   -1, -1, -2, -1, -2, -2, -3, -3,                 // sec
   -2, -3, -2, -3, -2, -5, -6, -2, -3, -2, -3, -3, // min
   -2, -3, -2, -3, -2, -5, -3, -3, -2,             // hour
   -1, -1, -2, -2, -4, -3, // day
   -6, -1, -2,             // week
   -2, -1, -2, -2, -2,     // month
   -1, -1, -2, -1, -1, -2, // year
   -1, -1, -2,             // 100year
   -1, -1, -2,             // 1000year
   -1, -1, -2,             // 10000year
   -1, -1, -2, -1, -1      // 100000year
};

// config
bool verbose;
double update_interval; // ms
TString ppp_up_com;
TString ppp_down_com;
double costs_per_minute;
double total_costs;
int debug; // 1==simulate load
int time_scale;

double min_wgrid_dy = 3;
double min_wgrid_dx = 3;

const char *options[] = {
   "#usage='Usage: %n [OPTIONS]\n\n"
     "this program is a gtk frontend to the simple ppp-up/down scripts with ppp load graph and lots of statistics",
   "#trailer='\n%n version %v *** (C) 1999 by Johannes Overmann\ncomments, bugs and suggestions welcome: %e\n%gpl'",
   "#stopat--", // stop option scanning after a -- parameter
   // options

   "name=timescale, type=int, save, default=21, lower=0, save, help=initial timescale, headline=config options:",
   "name=interval, type=double, save, default=1000.0, lower=50.0, upper=2000000000.0, help=update interval in milliseconds",
   "name=ppp-up-com, type=string, save, default=/etc/ppp/ppp-up, help=command to execute to connect",
   "name=ppp-down-com, type=string, save, default=/etc/ppp/ppp-down, help=command to execute to disconnect",
   "name=total-time, type=double, save, help=total amount of time connected in seconds",
   "name=total-costs, type=double, save, help=total amount of money spent",
   "name=costs-per-minute, type=double, save, default=0.05, help=online costs per minute",
   "name=debug, type=int, default=0, hide, help='1 == artificial ppp load graph'",
   "name=verbose, type=switch, char=v, save, help=verbose execution, headline=common options:",
   "EOL" // end of list
};


class TTimer {
 public:
   TTimer():started(0),elapsed(0) {}
   void start() {if(started==0) {started = time(0)-elapsed; elapsed=0;}}
   void stop() {if(started) {elapsed = time(0)-started; started=0;}}
   void reset(time_t set = 0) {if(started) started = time(0) - set; else elapsed = set;}
   void advance(time_t delta) {if(started) started -= delta; else elapsed += delta;}
   time_t current() const {if(started) return time(0)-started; else return elapsed;}
   bool isRunning() const {return started!=0;}
   bool isStopped() const {return started==0;}
   
 private:
   time_t started;
   time_t elapsed;
};



#define IP_NA_STR "n/a"
               
TArray<double> history_receive_bytes;
TArray<double> history_transmit_bytes;
TArray<double> history_throughput_bytes;
double last_receive_bytes = -1;
double last_transmit_bytes = -1;
double total_receive_bytes = 0;
double total_transmit_bytes = 0;
double total_throughput_bytes = 0;
double max_receive_bytes = 0;
double max_transmit_bytes = 0;
double max_throughput_bytes = 0;

bool ppp_up = false;
TString ppp_local_ip = IP_NA_STR;
TString ppp_remote_ip = IP_NA_STR;
TTimer ppp_uptime;
TTimer ppp_tot_uptime;
bool ignore_timer = false;

// gtk widgets
GtkWidget *wnd_main;
GtkWidget *lab_trans;
GtkWidget *lab_recv;
GtkWidget *lab_local;
GtkWidget *lab_remote;
GtkWidget *lab_current;
GtkWidget *lab_total;
GtkWidget *lab_scale;
GtkWidget *lab_about;
GtkWidget *lab_connect;
GtkWidget *lab_throughput;
GtkWidget *but_plus;
GtkWidget *but_minus;
GtkWidget *but_connect;
GtkWidget *dra_area;
/* pixmap as drawing area */
GdkPixmap *pixmap = NULL;
GdkColor c_transmit;
GdkColor c_receive;
GdkColor c_throughput;
GdkColor c_gstrong;
GdkColor c_gweak;
GdkColor c_gweak2;
GdkColor c_gweak3;
GdkColor c_gweak4;




TString bytes2str(double bytes, int num_digits = 3) {
   TString ret;
   const char *unit_str[] = {"b", "Kb", "Mb", "Gb", "Tb", "Pb", "Eb", "Zb", "Yb"};
   
   if(bytes < 1.0) {
      num_digits--;
      if(num_digits < 0) num_digits = 0;
      ret.sprintf("%.*fb", num_digits, bytes);
      return ret;
   }
   int unit = 0;   
   for(unit = 0; (bytes >= 1024.0) && (unit < 8); unit++) bytes /= 1024.0;
   if(bytes > 999.5) num_digits -= 4;
   else if(bytes >= 99.5) num_digits -= 3;
   else if(bytes >= 9.5) num_digits -= 2;
   else num_digits -= 1;
   if(num_digits < 0) num_digits = 0;
   ret.sprintf("%.*f%s", num_digits, bytes, unit_str[unit]);
   return ret;
}


TString time2str(time_t t) {
   time_t e = t;
   int s = e%60;  e/=60;
   int m = e%60;  e/=60;
   int h = e%24;  e/=24;
   int d = e%365; e/=365;
   int y = e;
   TString ret;
   if(t < 60*60) ret.sprintf("%d:%02d", m, s);
   else if(t < 24*60*60) ret.sprintf("%d:%02d:%02d", h, m, s);
   else if(t < 365*24*60*60) ret.sprintf("%dd %d:%02d:%02d", d, h, m, s);
   else ret.sprintf("%dy %dd %d:%02d:%02d", y, d, h, m, s);
   return ret;
}


TString sec2intervalstr(double sec) {
   TString r;
   
   if(sec < Tsec_min) r.sprintf("%gs", sec);
   else if(sec < Tsec_hour) r.sprintf("%gm", sec/Tsec_min);
   else if(sec < Tsec_day) r.sprintf("%gh", sec/Tsec_hour);
   else if(sec < Tsec_week) r.sprintf("%gd", sec/Tsec_day);
   else if(sec < Tsec_month) r.sprintf("%gw", sec/Tsec_week);
   else if(sec < Tsec_year)  r.sprintf("%gmo", sec/Tsec_month);
   else if(sec < 1000.0*Tsec_year)  r.sprintf("%gyr", sec/Tsec_year);
   else if(sec < 1000000.0*Tsec_year)  r.sprintf("%gmil", sec/Tsec_year/1000.0);
   else r.sprintf("%gumi", sec/Tsec_year/1000000.0);
   return r;
}


void draw_graph() {
   // x grid
   double x_grid[10];
   int x_grid_num = 0;
   int ts_pos = time_scale, ts_skip;
   double x_g;
   // y grid
   double y_wgrid = 0;
   double y_wgrid2 = 0;   
   double y_wgrid3 = 0;   
   double y_wgrid4 = 0;   
   int ymode = 10;   
   // main factors
   double y_v2p;
   double x_v2p;
   double xn_v2p;
   // misc
   double y_sgrid = 1.0;
   int had_grid_vals;
   int text_y;
   double text_val;
   int text_dy;
   int old_x = 0, old_y;
   int x, y=0, j, k;
   int max_dx, y_off, i;

   // pixmap to draw into already there?
   if(pixmap == 0) return;
   if(wnd_main == 0) return;
   
   // get size
   int dx, dy;
   gdk_window_get_size(pixmap, &dx, &dy);
   if(dx < 10) return;
   if(dy < 10) return;
   
   // clear pixmap, white
   gdk_draw_rectangle(pixmap, wnd_main->style->white_gc, TRUE,
		      0, 0, dx, dy);
   GdkGC *gc = gdk_gc_new(pixmap);
   
   // calc setup
   // there is a 32bit integer overflow in num_val for timescales >= 100years (for interval 1sec)
   // this will disappear when int is 64 bit which will be the case 
   // in 21xx I think 
   int num_vals = int(time_scale_tab[time_scale]/(update_interval/1000.0));
   double max_y = 1.0;
   int hist_len = history_transmit_bytes.num();
   if(hist_len < 1) goto free_gc;
   if(num_vals < 1) goto free_gc;
   i = hist_len - num_vals;
   if(i < 0) i = 0;
   for(; i < hist_len; i++) {
      if(history_throughput_bytes[i] > max_y) 
	max_y = history_throughput_bytes[i];
   }
   
   // main factors
   y_v2p = double(dy-1)/max_y;
   x_v2p = double(dx)/time_scale_tab[time_scale];
   xn_v2p = double(dx)/double(num_vals);
   
   // calc grid y   
   while(y_sgrid*y_v2p < dy*0.45) y_sgrid *= 1024;
   y_sgrid /= 1024;
   y_sgrid *= 1000;
   while(y_sgrid*y_v2p >= dy*0.45) y_sgrid /= 10;
   if(y_sgrid*y_v2p*2.0 < dy*0.45) {
      y_sgrid *= 2.0;
      ymode = 2;
   }
   if(y_sgrid*y_v2p*2.5 < dy*0.45) {
      y_sgrid *= 2.5;
      ymode = 5;
   }
   switch(ymode) {
    case 10:
    case 2:
      if(y_sgrid/2.0*y_v2p >= min_wgrid_dy) y_wgrid = y_sgrid/2.0;
      if(y_sgrid/10.0*y_v2p >= min_wgrid_dy) y_wgrid2 = y_sgrid/10.0;
      if(y_sgrid/20.0*y_v2p >= min_wgrid_dy) y_wgrid3 = y_sgrid/20.0;
      if(y_sgrid/100.0*y_v2p >= min_wgrid_dy) y_wgrid4 = y_sgrid/100.0;
      break;

    case 5:
      if(y_sgrid/5.0*y_v2p >= min_wgrid_dy) y_wgrid = y_sgrid/5.0;
      if(y_sgrid/10.0*y_v2p >= min_wgrid_dy) y_wgrid2 = y_sgrid/10.0;
      if(y_sgrid/50.0*y_v2p >= min_wgrid_dy) y_wgrid3 = y_sgrid/50.0;
      if(y_sgrid/100.0*y_v2p >= min_wgrid_dy) y_wgrid4 = y_sgrid/100.0;
      break;
   }
   // brighten weak grid colors
   y_off = 30;
   if(y_wgrid) while(y_wgrid4 == 0) {
      y_wgrid4 = y_wgrid3;
      y_wgrid3 = y_wgrid2;
      y_wgrid2 = y_wgrid;
      y_wgrid = 0;
      y_off -= 10;
   }
   
   // calc grid x
   for(x_grid_num = 0; x_grid_num < 5; x_grid_num++) {
      // get next possible value
      ts_skip = time_grid_skip[ts_pos];
      if(ts_skip == 0) break;
      ts_pos += ts_skip;
      x_g = time_scale_tab[ts_pos];
      // check if it is valid
      if(x_g*x_v2p < min_wgrid_dx) break;
      // store it
      x_grid[x_grid_num] = x_g;
   }
   had_grid_vals = x_grid_num;
   x_grid_num--;
   
   max_dx = dx - 2;

   // draw weak4 grid
   if(y_wgrid4) {
      gdk_gc_set_foreground(gc, &c_gweak4);
      for(i = 1;; i++) {      
	 int y = int(y_wgrid4*y_v2p*i);
	 if(y >= dy) break;
	 gdk_draw_line(pixmap, gc, y_off, dy-y, dx-1, dy-y);      
      }
   }
   if(x_grid_num>=0) {
      gdk_gc_set_foreground(gc, &c_gweak4);
      for(i = 1;; i++) {      
	 int x = int(x_grid[x_grid_num]*x_v2p*i);
	 if(x >= max_dx) break;
	 gdk_draw_line(pixmap, gc, dx-x, y_off, dx-x, dy-1);
      }      
   }
   x_grid_num--;
   y_off -= 10;
   if(y_off < 0) y_off = 0;
      
   // draw weak3 grid
   if(y_wgrid3) {
      gdk_gc_set_foreground(gc, &c_gweak3);
      for(i = 1;; i++) {      
	 int y = int(y_wgrid3*y_v2p*i);
	 if(y >= dy) break;
	 gdk_draw_line(pixmap, gc, y_off, dy-y, dx-1, dy-y);      
      }
   }
   if(x_grid_num>=0) {
      gdk_gc_set_foreground(gc, &c_gweak3);
      for(i = 1;; i++) {      
	 int x = int(x_grid[x_grid_num]*x_v2p*i);
	 if(x >= max_dx) break;
	 gdk_draw_line(pixmap, gc, dx-x, y_off, dx-x, dy-1);
      }      
   }
   x_grid_num--;
   y_off -= 10;
   if(y_off < 0) y_off = 0;
      
   // draw weak2 grid
   if(y_wgrid2) {
      gdk_gc_set_foreground(gc, &c_gweak2);
      for(i = 1;; i++) {      
	 int y = int(y_wgrid2*y_v2p*i);
	 if(y >= dy) break;
	 gdk_draw_line(pixmap, gc, y_off, dy-y, dx-1, dy-y);
      }
   }
   if(x_grid_num>=0) {
      gdk_gc_set_foreground(gc, &c_gweak2);
      for(i = 1;; i++) {      
	 int x = int(x_grid[x_grid_num]*x_v2p*i);
	 if(x >= max_dx) break;
	 gdk_draw_line(pixmap, gc, dx-x, y_off, dx-x, dy-1);
      }      
   }
   x_grid_num--;
      
   // draw weak grid
   if(y_wgrid) {
      gdk_gc_set_foreground(gc, &c_gweak);
      for(i = 1;; i++) {      
	 int y = int(y_wgrid*y_v2p*i);
	 if(y >= dy) break;
	 gdk_draw_line(pixmap, gc, 0, dy-y, dx-1, dy-y);      
      }
   }   
   if(x_grid_num>=0) {
      gdk_gc_set_foreground(gc, &c_gweak);
      for(i = 1;; i++) {      
	 int x = int(x_grid[x_grid_num]*x_v2p*i);
	 if(x >= max_dx) break;
	 gdk_draw_line(pixmap, gc, dx-x, 0, dx-x, dy-1);
      }      
   }
   x_grid_num--;
      
   // draw strong grid
   gdk_gc_set_foreground(gc, &c_gstrong);
   text_dy = gdk_text_height(wnd_main->style->font, "08M.", 4);
   for(i = 1;; i++) {      
      int y = int(y_sgrid*y_v2p*i);
      if(y >= dy) break;
      gdk_draw_line(pixmap, gc, 0, dy-y, dx-1, dy-y);
      text_y = dy - y;
      text_val = y_sgrid*i;
      // print text
      TString str = bytes2str(text_val, 2);
      str += "/s";
      text_y -= 2;
      if(text_y > 2*(text_dy + 2)) 
	gdk_draw_text(pixmap, wnd_main->style->font,
		      wnd_main->style->black_gc,
		      2, text_y, str, str.len());
   }   
   if(x_grid_num>=0) {
      gdk_gc_set_foreground(gc, &c_gstrong);
      for(i = 1;; i++) {      
	 int x = int(x_grid[x_grid_num]*x_v2p*i);
	 if(x >= max_dx) break;
	 gdk_draw_line(pixmap, gc, dx-x, 0, dx-x, dy-1);
      }      
   }
   if(had_grid_vals) {
      gdk_gc_set_foreground(gc, &c_gstrong);
      for(i = 1;; i++) {      
	 int x = int(x_grid[0]*x_v2p*i);
	 if(x > dx) break;
	 TString str = sec2intervalstr(x_grid[0]*i);
	 gdk_draw_text(pixmap, wnd_main->style->font,
		       wnd_main->style->black_gc,
		       (dx-x) + 2, text_dy + 2, str, str.len());
	 
      }
   }
   
   // draw graphs
   gdk_gc_set_foreground(gc, &c_throughput);
   for(k = 0; k < 3; k++) {
      switch(k) {
       case 0:
	 gdk_gc_set_foreground(gc, &c_transmit);
	 break;
       case 1:
	 gdk_gc_set_foreground(gc, &c_receive);
	 break;
       case 2:
	 gdk_gc_set_foreground(gc, &c_throughput);
	 break;	    
      }
      old_y = -1;
      for(i = hist_len-1, j = 0; (i>=0) && (j <= num_vals); i--, j++) {
	 x = int(double(j)*xn_v2p);
	 switch(k) {
	  case 0:
	    y = int(double(history_transmit_bytes[i])*y_v2p);
	    break;
	  case 1:
	    y = int(double(history_receive_bytes[i])*y_v2p);
	    break;
	  case 2:
	    y = int(double(history_throughput_bytes[i])*y_v2p);
	    break;	    
	 }
	 if(old_y >= 0) gdk_draw_line(pixmap, gc, dx-x, dy-y-1, dx-old_x, dy-old_y-1);
	 old_x = x;
	 old_y = y;            
      }
   }

free_gc:   
   // cleanup
   gdk_gc_destroy(gc);

   // redraw widget
   gtk_widget_draw(dra_area, NULL);   
}


double costs(time_t sec) {
   sec = (sec+59)/60;
   return double(sec)*costs_per_minute;
}


void doDiff(double &last, double &cur) {
   double t = last;
   static double fact = 1.0;
   last = cur;
   if(t < 0.0) cur = 0.0;
   else cur -= t;
   if(cur < 0.0) cur = 0.0;
   if(debug&1) {
      cur = fact;
      fact *= 1.02;
   }
}


// called everytime the pppd goes up or down
void new_ppp_state() {
   if(ppp_up) { // pppd up
      gtk_label_set(GTK_LABEL(lab_connect), "Disconnect");
   } else { // ppp down
      gtk_label_set(GTK_LABEL(lab_connect), "Connect");
   }   
   gtk_label_set(GTK_LABEL(lab_local), ppp_local_ip);
   gtk_label_set(GTK_LABEL(lab_remote), ppp_remote_ip);
}


// called every time the time scale changes
void new_time_scale() {
   gtk_label_set(GTK_LABEL(lab_scale), sec2intervalstr(time_scale_tab[time_scale]));
   draw_graph();
}


// reset ppp uptime
void reset_ppp_uptime() {
   ppp_tot_uptime.advance(ppp_uptime.current());
   total_costs += costs(ppp_uptime.current());
   ppp_uptime.reset();
}


gint progress_timeout( gpointer )
{
   static int first_time = 1;
   char buf[80];
   
   // application going down?
   if(ignore_timer) return false;
   
   // read from ifconfig
   FILE *f = popen("/sbin/ifconfig", "r");
   if(f) { 
      TString wline;      
      bool cur_ppp_up = false;
      // scan to ppp0
      for(;;) {
	 wline.readLine(f);
	 wline.cropSpaceEnd();
	 if(feof(f)) break;
	 if(wline.hasPrefix("ppp0")) break;
      }
      if(wline.hasPrefix("ppp0")) {
	 cur_ppp_up = true;
	 wline.readLine(f);
	 wline.translateChar(':', ' ');
	 wline.collapseSpace();
	 wline.cropSpace();
	 TArray<TString> a = split(wline, " ");
	 ppp_local_ip = a[2];
	 ppp_remote_ip = a[4];
      }
      // state changed ?
      if((cur_ppp_up != ppp_up) || first_time) {
	 if(cur_ppp_up) {
	    // pppd up
	    reset_ppp_uptime();
	    ppp_uptime.start();
	 } else {
	    // pppd down
	    ppp_local_ip = IP_NA_STR;
	    ppp_remote_ip = IP_NA_STR;
	    ppp_uptime.stop();
	 }
	 ppp_up = cur_ppp_up;
	 // update state display
	 new_ppp_state();
	 first_time = 0;
      }      
      pclose(f);
   }
   // read from /proc/net/dev
   f = fopen("/proc/net/dev", "rt");
   if(f) {
      TString wline;      
      // scan to ppp0:
      for(;;) {
	 wline.readLine(f);
	 wline.cropSpaceEnd();
	 if(feof(f)) break;
	 if(wline.search("ppp0:")) break;
      }      
      double rec = 0.0;
      double tra = 0.0;
      if(wline.search("ppp0:")) {
	 wline.translateChar(':', ' ');
	 wline.collapseSpace();
	 wline.cropSpace();
	 TArray<TString> a = split(wline, " ");
	 a[1].toDouble(rec);	 
	 a[9].toDouble(tra);
      }
      doDiff(last_receive_bytes, rec);
      history_receive_bytes += rec;
      total_receive_bytes += rec;
      if(rec > max_receive_bytes) max_receive_bytes = rec;
      doDiff(last_transmit_bytes, tra);
      history_transmit_bytes += tra;
      total_transmit_bytes += tra;
      if(tra > max_transmit_bytes) max_transmit_bytes = tra;
      double tr = rec+tra;
      history_throughput_bytes += tr;
      total_throughput_bytes += tr;
      if(tr > max_throughput_bytes) max_throughput_bytes = tr;	 
      fclose(f);
   }

   // update graph
   draw_graph();
   
   // time display
   TString str;
   str.sprintf("%.2f", costs(ppp_uptime.current()));
   gtk_label_set(GTK_LABEL(lab_current), time2str(ppp_uptime.current()) + " / " + str);
   str.sprintf("%.2f", costs(ppp_uptime.current()) + total_costs);
   gtk_label_set(GTK_LABEL(lab_total), time2str(ppp_uptime.current() + ppp_tot_uptime.current()) + " / " + str);
   
   // rec/trans
   int ndig = 4;
   sprintf(buf, "%s / %s/s", *bytes2str(total_transmit_bytes, ndig), *bytes2str(max_transmit_bytes, ndig));
   gtk_label_set(GTK_LABEL(lab_trans), buf);
   sprintf(buf, "%s / %s/s", *bytes2str(total_receive_bytes, ndig), *bytes2str(max_receive_bytes, ndig));
   gtk_label_set(GTK_LABEL(lab_recv), buf);
   sprintf(buf, "%s / %s/s", *bytes2str(total_throughput_bytes, ndig), *bytes2str(max_throughput_bytes, ndig));
   gtk_label_set(GTK_LABEL(lab_throughput), buf);
   
   // we want to get called again
   return true; 
} 



void create_window(const TString& version) {
   GtkWidget *hbox1;
   GtkWidget *frame8;
   GtkWidget *vbox1;
   GtkWidget *hbox4;
   GtkWidget *frame3;
   GtkWidget *frame4;
   GtkWidget *hbox3;
   GtkWidget *frame1;
   GtkWidget *frame2;
   GtkWidget *hbox2;
   GtkWidget *frame6;
   GtkWidget *frame7;
   GtkWidget *hbox5;
   GtkWidget *frame5;
   GtkWidget *frame9;
   GtkWidget *hbox6;
   GtkWidget *hbox7;

   wnd_main = gtk_window_new(GTK_WINDOW_TOPLEVEL);
   gtk_container_set_border_width(GTK_CONTAINER(wnd_main), 5);
   gtk_window_set_title(GTK_WINDOW(wnd_main), "SimplePPP V" + version);
   gtk_widget_set_usize(wnd_main, 300, 43); // min size
   gtk_window_set_default_size(GTK_WINDOW(wnd_main), 600, 200); // default size
   
   hbox1 = gtk_hbox_new(FALSE, 5);
   gtk_container_add(GTK_CONTAINER(wnd_main), hbox1);

   frame8 = gtk_frame_new(NULL);
   gtk_box_pack_start(GTK_BOX(hbox1), frame8, TRUE, TRUE, 0);
   gtk_frame_set_shadow_type(GTK_FRAME(frame8), GTK_SHADOW_IN);
   
   dra_area = gtk_drawing_area_new();
   gtk_container_add(GTK_CONTAINER(frame8), dra_area);  

   vbox1 = gtk_vbox_new(FALSE, 3);
   gtk_box_pack_start(GTK_BOX(hbox1), vbox1, FALSE, FALSE, 0);
   
   hbox4 = gtk_hbox_new(TRUE, 5);
   gtk_box_pack_start(GTK_BOX(vbox1), hbox4, FALSE, FALSE, 0);

   frame3 = gtk_frame_new("Transmitted / Max");
   gtk_box_pack_start(GTK_BOX(hbox4), frame3, TRUE, TRUE, 0);

   lab_trans = gtk_label_new("0 / 0.0");
   gtk_container_add(GTK_CONTAINER(frame3), lab_trans);

   frame4 = gtk_frame_new("Received / Max");
   gtk_box_pack_start(GTK_BOX(hbox4), frame4, TRUE, TRUE, 0);
   
   lab_recv = gtk_label_new("0 / 0.0");
   gtk_container_add(GTK_CONTAINER(frame4), lab_recv);

   hbox7 = gtk_hbox_new(TRUE, 5);
   gtk_box_pack_start(GTK_BOX(vbox1), hbox7, FALSE, FALSE, 0);

   hbox3 = gtk_hbox_new(TRUE, 5);
   gtk_box_pack_start(GTK_BOX(vbox1), hbox3, FALSE, FALSE, 0);

   frame1 = gtk_frame_new("Local IP Address");
   gtk_box_pack_start(GTK_BOX(hbox3), frame1, TRUE, TRUE, 0);

   lab_local = gtk_label_new("n/a");
   gtk_container_add(GTK_CONTAINER(frame1), lab_local);

   frame2 = gtk_frame_new("Remote IP Address");
   gtk_box_pack_start(GTK_BOX(hbox3), frame2, TRUE, TRUE, 0);

   lab_remote = gtk_label_new("n/a");
   gtk_container_add(GTK_CONTAINER(frame2), lab_remote);

   hbox2 = gtk_hbox_new(TRUE, 5);
   gtk_box_pack_start(GTK_BOX(vbox1), hbox2, FALSE, FALSE, 0);

   frame6 = gtk_frame_new("Current Time / Costs");
   gtk_box_pack_start(GTK_BOX(hbox2), frame6, TRUE, TRUE, 0);

   lab_current = gtk_label_new("0:00 / 0.00");
   gtk_container_add(GTK_CONTAINER(frame6), lab_current);

   frame7 = gtk_frame_new("Total Time / Costs");
   gtk_box_pack_start(GTK_BOX(hbox2), frame7, TRUE, TRUE, 0);

   lab_total = gtk_label_new("0:00 / 0.00");
   gtk_container_add(GTK_CONTAINER(frame7), lab_total);

   hbox5 = gtk_hbox_new(TRUE, 5);
   gtk_box_pack_start(GTK_BOX(vbox1), hbox5, FALSE, FALSE, 0);

   frame5 = gtk_frame_new("Time Scale");
   gtk_box_pack_start(GTK_BOX(hbox5), frame5, TRUE, TRUE, 0);

   frame9 = gtk_frame_new("Throughput / Max");
   gtk_box_pack_start(GTK_BOX(hbox7), frame9, TRUE, TRUE, 0);

   hbox6 = gtk_hbox_new(FALSE, 0);
   gtk_container_add(GTK_CONTAINER(frame5), hbox6);

   but_plus = gtk_button_new_with_label(" + ");
   gtk_box_pack_start(GTK_BOX(hbox6), but_plus, FALSE, FALSE, 0);
   gtk_widget_set_usize(but_plus, 28, 28);
   gtk_container_set_border_width(GTK_CONTAINER(but_plus), 3);

   lab_scale = gtk_label_new("1m");
   gtk_box_pack_start(GTK_BOX(hbox6), lab_scale, TRUE, TRUE, 0);

   but_minus = gtk_button_new_with_label(" - ");
   gtk_box_pack_start(GTK_BOX(hbox6), but_minus, FALSE, FALSE, 0);
   gtk_widget_set_usize(but_minus, 28, 28);
   gtk_container_set_border_width(GTK_CONTAINER(but_minus), 3);

   lab_connect = gtk_label_new("con");   
   but_connect = gtk_button_new();
   gtk_container_add(GTK_CONTAINER(but_connect), lab_connect);
   gtk_box_pack_start(GTK_BOX(hbox5), but_connect, TRUE, TRUE, 0);

   lab_about = gtk_label_new("SimplePPP V" + version);
   gtk_box_pack_start(GTK_BOX(hbox7), lab_about, TRUE, TRUE, 0);
   
   lab_throughput = gtk_label_new("0.0 / 0.0");
   gtk_container_add(GTK_CONTAINER(frame9), lab_throughput);
}


void but_click_connect(GtkWidget *, gpointer) {
   if(ppp_up) system(ppp_down_com);
   else system(ppp_up_com);
}


void but_click_plus(GtkWidget *, gpointer) {
   if(time_scale_tab[time_scale+1]) {
      if(time_scale_tab[time_scale+1]/(update_interval/1000.0) <= (1024.0*1024.0*1024.0*2.0-1.0)) {
	 time_scale++;
	 new_time_scale();
      }
   }
}


void but_click_minus(GtkWidget *, gpointer) {
   if(time_scale > 0) {
      if(time_scale_tab[time_scale-1]/(update_interval/1000.0) >= 1.0) {
	 time_scale--;
	 new_time_scale();
      }
   }
}


// create a new pixmap of the appropriate size 
static gint
configure_event(GtkWidget *widget, GdkEventConfigure *) {
   if(pixmap)
     gdk_pixmap_unref(pixmap);
   pixmap = gdk_pixmap_new(widget->window, widget->allocation.width,
			   widget->allocation.height, -1);
   draw_graph();
   return TRUE;
}


// redraw the screen from the backing pixmap 
static gint expose_event(GtkWidget *widget, GdkEventExpose *event)
{
  gdk_draw_pixmap(widget->window,
		  widget->style->fg_gc[GTK_WIDGET_STATE(widget)],
		  pixmap,
		  event->area.x, event->area.y,
		  event->area.x, event->area.y,
		  event->area.width, event->area.height);
  return FALSE;
}


void quit() {
   ignore_timer = true;
   gtk_main_quit();
}


int main(int argc, char *argv[]) {
   // initialize gtk and parse gtk commands
   gtk_init (&argc, &argv);
   
   // get parameters
   TAppConfig ac(options, "options", argc, argv, "SIMPLEPPPOPT", "simpleppprc", VERSION);
   // setup
   verbose = ac("verbose");
   ppp_up_com = ac.getString("ppp-up-com");
   ppp_down_com = ac.getString("ppp-down-com");
   update_interval = ac.getDouble("interval");
   costs_per_minute = ac.getDouble("costs-per-minute");
   total_costs = ac.getDouble("total-costs");
   ppp_tot_uptime.reset(int(ac.getDouble("total-time")));
   time_scale = ac.getInt("timescale");
   debug = ac.getInt("debug");
   
   // check args
   if(ac.numParam())
     userError("no arguments allowed\n");

   // create main window
   create_window(ac.getString("application-version"));
   
   // allocate colors
   GdkColormap *cmap = gdk_colormap_get_system();      
   c_transmit.red = 0xffff;
   c_transmit.green = 0;
   c_transmit.blue = 0;
   if(!gdk_color_alloc(cmap, &c_transmit))
     userError("can't allocate color!\n!");
   c_receive.red = 0;
   c_receive.green = 0x5000;
   c_receive.blue = 0;
   if(!gdk_color_alloc(cmap, &c_receive))
     userError("can't allocate color!\n!");
   c_throughput.red = 0;
   c_throughput.green = 0;
   c_throughput.blue = 0xa000;
   if(!gdk_color_alloc(cmap, &c_throughput))
     userError("can't allocate color!\n!");
   c_gstrong.red = 0x4000;
   c_gstrong.green = 0x4000;
   c_gstrong.blue = 0x4000;
   if(!gdk_color_alloc(cmap, &c_gstrong))
     userError("can't allocate color!\n!");
   c_gweak.red = 0xa000;
   c_gweak.green = 0xa000;
   c_gweak.blue = 0xa000;
   if(!gdk_color_alloc(cmap, &c_gweak))
     userError("can't allocate color!\n!");
   c_gweak2.red = 0xc000;
   c_gweak2.green = 0xc000;
   c_gweak2.blue = 0xc000;
   if(!gdk_color_alloc(cmap, &c_gweak2))
     userError("can't allocate color!\n!");
   c_gweak3.red = 0xe000;
   c_gweak3.green = 0xe000;
   c_gweak3.blue = 0xe000;
   if(!gdk_color_alloc(cmap, &c_gweak3))
     userError("can't allocate color!\n!");
   c_gweak4.red = 0xf000;
   c_gweak4.green = 0xf000;
   c_gweak4.blue = 0xf000;
   if(!gdk_color_alloc(cmap, &c_gweak4))
     userError("can't allocate color!\n!");
   
   // activate close gadget
   gtk_signal_connect(GTK_OBJECT(wnd_main), "destroy",
		       GTK_SIGNAL_FUNC(quit), NULL);
   
   // Signals used to handle backing pixmap    
   gtk_signal_connect(GTK_OBJECT(dra_area), "expose_event",
		      (GtkSignalFunc) expose_event, NULL);
   gtk_signal_connect(GTK_OBJECT(dra_area),"configure_event",
		      (GtkSignalFunc) configure_event, NULL);
   
   // Event signals   
   gtk_widget_set_events(dra_area, GDK_EXPOSURE_MASK);

   // button clicked events
   gtk_signal_connect(GTK_OBJECT(but_connect), "clicked",
		      (GtkSignalFunc)but_click_connect, NULL);
   gtk_signal_connect(GTK_OBJECT(but_plus), "clicked",
		      (GtkSignalFunc)but_click_plus, NULL);
   gtk_signal_connect(GTK_OBJECT(but_minus), "clicked",
		      (GtkSignalFunc)but_click_minus, NULL);
   
   // show main window
   gtk_widget_show_all(wnd_main);
   
   // Add a timer callback to update the value of the progress bar
   gtk_timeout_add(int(update_interval), progress_timeout, NULL);

   //
   new_ppp_state();
   new_time_scale();
   progress_timeout(0);
   
   // gtk main loop
   gtk_main();
   ignore_timer = true;

   // save data
   reset_ppp_uptime();
   ac.setValue("total-costs", total_costs);
   ac.setValue("total-time", double(ppp_tot_uptime.current()));
   ac.setValue("timescale", time_scale);
   TString savename;
   ac.save(savename);
   
   // normal exit
   return 0;
}


